import numpy as np
import os 
from scipy import optimize

class Parameter:
    def __init__(self, value, name=''):
        self.value = value
        self.name = name

    def set(self, value):
        self.value = value

    def __call__(self):
        return self.value

def fit1d(x, y, fitmethod, *arg, **kw):

    do_print = kw.pop('do_print', False)
    ret = kw.pop('ret', True)
    fixed = kw.pop('fixed', [])
    VERBOSE= kw.pop('VERBOSE',False)

    y_err = kw.pop ('y_err', None)
    if y_err is not None:
        if (len(y_err) != len(y)):
            print('Data and error arrays have non-matching lengths!')
            y_err = None

    if fitmethod != None:
        p0, fitfunc, fitfunc_str = fitmethod(*arg)
    else:
        p0 = kw.pop('p0')
        fitfunc = kw.pop('fitfunc')
        fitfunc_str = kw.pop('fitfunc_str', '')        
    
    fixedp = []
    for i,p in enumerate(p0):
        if i in fixed:
            fixedp.append(p)
    for p in fixedp:
        p0.remove(p)

    def f(params):
        i = 0
        for p in p0:
            p.set(params[i])
            i += 1

        if y_err is not None:
            return ((y-fitfunc(x))/(y_err))
        else:
            return y - fitfunc(x)

    if x is None: x = arange(y.shape[0])
    p = [param() for param in p0]

    p1, cov, info, mesg, success = optimize.leastsq(f, p, full_output=True, maxfev=len(x)*100)

    if not success or (cov == None if (not isinstance(cov,(list, tuple, np.ndarray))) else False):
        success = False
        if VERBOSE:
            print('ERROR: Fit did not converge !')
            print('reason: ',mesg)
    
    if y_err is not None:
        result = result_dict(p1, cov, info, mesg, success, x, y, p0, 
                fitfunc, fitfunc_str, y_err=y_err)
    else:
        result = result_dict(p1, cov, info, mesg, success, x, y, p0, 
                fitfunc, fitfunc_str)
    
    if do_print and success:
        print_fit_result(result)

    if ret:
        return result

def result_dict(p1, cov, info, mesg, success, x, y, p0, fitfunc, fitfunc_str, y_err=None):
    chisq = 1.
    dof = 1.
    gof=1.
    error_dict = {}
    error_list = []
    params_dict = {}

    if success:
        chisq = sum(info['fvec']*info['fvec'])

        # gof =  sum(info['fvec']*info['fvec']/(fitfunc(x)*fitfunc(x)))
        dof = float(len(y)-len(p0))

        for i,pmin in enumerate(p1):
            if y_err is not None:
                error_dict[p0[i].name] = np.sqrt(cov[i,i]) 
                error_list.append(np.sqrt(cov[i,i])*np.sqrt(chisq/dof))
                params_dict[p0[i].name] = pmin
            else:
                error_dict[p0[i].name] = np.sqrt(cov[i,i])*np.sqrt(chisq/dof)
                error_list.append(np.sqrt(cov[i,i])*np.sqrt(chisq/dof))
                params_dict[p0[i].name] = pmin

    result = {
        'success' : success,
        'params' : p1,
        'params_dict' : params_dict,
        'chisq': chisq,
        'dof': dof,
        'gof': gof,
        'residuals_rms': np.sqrt(chisq/dof),
        'reduced_chisq': chisq/dof,
        'error' : error_list,
        'error_dict' : error_dict, 
        'cov' : cov,
        'p0' : p0,
        'fitfunc' : fitfunc,
        'fitfunc_str' : fitfunc_str,
        'x' : x,
        'y' : y,
        'y_err': y_err
        }
    
    return result

def fit_SEDOR_insens_non_decay(g_delta, g_eps, g_f,g_phi=0.,*arg):
    fitfunc_str = '0.5 + 0.5*( delta* (1-eps) )*np.cos(2*pi*f*t) + 0.5*delta*eps'

    f     = Parameter(g_f, 'f')
    eps   = Parameter(g_eps, 'eps')
    delta = Parameter(g_delta, 'delta')
    phi   = Parameter(g_phi, 'phi')

    p0 = [delta,eps, f,phi]

    eta = 1

    def fitfunc(t):
        return 0.5 + 0.5*delta()*(1-eps())*np.cos(2*np.pi*f()*t + phi()*np.pi/180.) + 0.5*delta()*eps()

    return p0, fitfunc, fitfunc_str

def fit_SEDOR_insens_decay(g_delta, g_eps, g_f, g_T, g_n, *arg):
    fitfunc_str = 'return 0.5*( 1 + delta*(1-eps)*exp(-(t/T)**n)*cos(2pi*f*t) + delta*eps*exp(-(t/T)**n) )'

    f     = Parameter(g_f, 'f')
    eps   = Parameter(g_eps, 'eps')
    delta = Parameter(g_delta, 'delta')
    T     = Parameter(g_T, 'T')
    n     = Parameter(g_n, 'n')

    p0 = [delta, eps, f, T, n]

    def fitfunc(t):
        return 0.5*(1+delta()*(1-eps()) *np.cos(2*np.pi*f()*t)*np.exp(-1*(t/T())**n())+delta()*eps()*np.exp(-1*(t/T())**n()) )    

    return p0, fitfunc, fitfunc_str

def fit_parabole(g_o, g_A, g_c, *arg):
    """'o + A * (x-c)**2'
    """
    fitfunc_str = 'o + A * (x-c)**2'

    o = Parameter(g_o, 'o')
    A = Parameter(g_A, 'A')
    c = Parameter(g_c, 'c')
    p0 = [o, A, c]

    def fitfunc(x):
        return o() + A() * (x-c())**2

    return p0, fitfunc, fitfunc_str

def fit_cos(g_f, g_a, g_A, g_phi, *arg):
    """A * cos(2pi * (f*x + phi/360) ) + a
    """
    fitfunc_str = 'A * cos(2pi * (f*x + phi/360) ) + a'

    f = Parameter(g_f, 'f')
    a = Parameter(g_a, 'a')
    A = Parameter(g_A, 'A')
    phi = Parameter(g_phi, 'phi')

    p0 = [f, a, A, phi]

    def fitfunc(x):
        return a() + A() * np.cos(2*np.pi*( f()*x + phi()/360.))

    return p0, fitfunc, fitfunc_str

def fit_general_exponential(g_a, g_A, g_x0, g_T, g_n):
    fitfunc_str = 'a + A * exp(-((x-x0)/T )**n)'

    a = Parameter(g_a, 'a')
    A = Parameter(g_A, 'A')
    x0 = Parameter(g_x0, 'x0')
    T = Parameter(g_T, 'T')
    n = Parameter(g_n, 'n')

    p0 = [a, A, x0, T, n]

    def fitfunc(x):
        return a() + A() * np.exp(-(x-x0())**n()/(T()**n()))
    return p0, fitfunc, fitfunc_str


def fit_gauss(g_a, g_A, g_x0, g_sigma):
    fitfunc_str = 'a + A * exp(-(x-x0)**2/(2*sigma**2))'

    a = Parameter(g_a, 'a')
    x0 = Parameter(g_x0, 'x0')
    A = Parameter(g_A, 'A')
    sigma = Parameter(g_sigma, 'sigma')

    p0 = [a, A, x0, sigma]

    def fitfunc(x):
        return a() + A() * np.exp(-(x-x0())**2/(2*sigma()**2))
    return p0, fitfunc, fitfunc_str


def fit_double_gauss(g_a, g_A1, g_x01, g_sigma1,g_A2, g_x02, g_sigma2):
    fitfunc_str = 'a + A1 * exp(-(x-x01)**2/(2*sigma1**2))+ A2* exp(-(x-x02)**2/(2*sigma2**2))'

    a = Parameter(g_a, 'a')
    x01 = Parameter(g_x01, 'x01')
    A1 = Parameter(g_A1, 'A1')
    sigma1 = Parameter(g_sigma1, 'sigma1')
    x02 = Parameter(g_x02, 'x02')
    A2 = Parameter(g_A2, 'A2')
    sigma2 = Parameter(g_sigma2, 'sigma2')

    p0 = [a, A1, x01, sigma1, A2, x02, sigma2]

    def fitfunc(x):
        return a() + A1() * np.exp(-(x-x01())**2/(2*sigma1()**2)) + A2() * np.exp(-(x-x02())**2/(2*sigma2()**2))
    return p0, fitfunc, fitfunc_str



def normalise_data(y_vals, y_err_vals, normalisation_fit_params, normalisation_fit_errs, probe_spins, carbon_nr):
    
    if probe_spins[carbon_nr][0] == None:
        norm = normalisation_fit_params[carbon_nr]
        norm_err = normalisation_fit_errs[carbon_nr]
        
        expec_Z = (2*y_vals - 1)/norm
        expec_Z_err = np.sqrt(4*y_err_vals**2/norm**2 + ((1-2*y_vals)**2 * norm_err**2)/norm**4)
        
    elif probe_spins[carbon_nr][2] == 'decay':
    
        delta = normalisation_fit_params[carbon_nr]['delta']
        delta_err = normalisation_fit_errs[carbon_nr]['delta']

        eps = normalisation_fit_params[carbon_nr]['eps']
        eps_err = normalisation_fit_errs[carbon_nr]['eps']

        T = normalisation_fit_params[carbon_nr]['T']
        T_err = normalisation_fit_errs[carbon_nr]['T']

        n = normalisation_fit_params[carbon_nr]['n']
        n_err = normalisation_fit_errs[carbon_nr]['n']

        f = normalisation_fit_params[carbon_nr]['f']
        f_err = normalisation_fit_errs[carbon_nr]['f']

        expec_Z = (2*y_vals - 1) / (delta*np.sin(f*np.pi)*np.exp(-(0.5/T)**n)*(1-eps))

        func1 = np.exp(2**(1-n)*T**(-n))/(delta**4*(eps - 1)**4)
        func2 = (1/np.sin(f*np.pi))**2*(4**n*(delta**2*eps_err**2 + delta_err**2*(eps - 1)**2) + 4**n*np.pi**2*delta**2*f_err**2*(eps - 1)**2 *(1/np.tan(f*np.pi))**2 + T**(-2*(1+n))*delta**2*(eps-1)**2*(n**2*T_err**2 + T**2*n_err**2*np.log(2*T)**2) )
        func3 = 4**(-n)*(1-2*y_vals)**2*abs(func2) + 4*delta**2*y_err_vals**2*(eps - 1)**2*(1/np.sin(f*np.pi)**2)

        expec_Z_err = np.sqrt(func1*func3)

    else:
    
        delta = normalisation_fit_params[carbon_nr]['delta']
        delta_err = normalisation_fit_errs[carbon_nr]['delta']

        eps = normalisation_fit_params[carbon_nr]['eps']
        eps_err = normalisation_fit_errs[carbon_nr]['eps']

        f = normalisation_fit_params[carbon_nr]['f']
        f_err = normalisation_fit_errs[carbon_nr]['f']

        expec_Z = (2*y_vals - 1)/(delta*np.sin(f*np.pi)*(1-eps))

        num = (1 - 2*y_vals)**2 * abs((delta**2*eps_err**2 + delta_err**2*(eps - 1)**2 + np.pi**2*delta**2*f_err**2*(eps - 1)**2*(1/np.tan(f*np.pi))**2)*(1/np.sin(f*np.pi))**2) + 4*delta**2*y_err_vals**2*(eps - 1)**2*(1/np.sin(f*np.pi))**2
        den = delta**4*(eps - 1)**4

        expec_Z_err = np.sqrt( num/den )
    
    return expec_Z, expec_Z_err




def fit_general_decaying_cos(g_f, g_a, g_A, g_phi,g_t, g_n, *arg):
    fitfunc_str = 'A *exp(-x/t) cos(2pi * (f*x + phi/360) ) + a'

    f = Parameter(g_f, 'f')
    a = Parameter(g_a, 'a')
    A = Parameter(g_A, 'A')
    phi = Parameter(g_phi, 'phi')
    t   = Parameter(g_t, 't')
    n = Parameter(g_n, 'n')
    p0 = [f, a, A,phi,t, n]

    def fitfunc(x):
        return a() + A()*np.exp(-(x/t())**n()) * np.cos(2*np.pi*( f()*x + phi()/360.))

    return p0, fitfunc, fitfunc_str

def fit_general_decaying_2cos(g_avg, g_t1, g_t2, g_A, g_f_a, g_phi_a, g_n_a, g_B, g_f_b, g_phi_b, g_n_b, *arg):

    fitfunc_str = 'avg + C * exp(-(x/t)**2) [ A*cos(2pi * (fa*x + phi_a/360) )+ B*cos(2pi * (f_b*x + phi_b/360)))/2 ]'

    avg = Parameter(g_avg, 'avg')

    t1   = Parameter(g_t1, 't1')
    t2   = Parameter(g_t2, 't2')

    A = Parameter(g_A, 'A')
    f_a = Parameter(g_f_a, 'f_a')
    phi_a = Parameter(g_phi_a, 'phi_a')
    n_a = Parameter(g_n_a, 'n_a')

    B = Parameter(g_B, 'B')
    f_b = Parameter(g_f_b, 'f_b')
    phi_b = Parameter(g_phi_b, 'phi_b')
    n_b = Parameter(g_n_b, 'n_b')

    p0 = [avg, t1, t2, A,f_a,phi_a,n_a,B,f_b,phi_b,n_b]

    def fitfunc(x):
        return avg() + np.exp(-(x/t1())**n_a()) *  A()*np.cos(2*np.pi*( f_a()*x + phi_a()/360.)) + np.exp(-(x/t2())**n_b()) * B() * np.cos(2*np.pi*( f_b()*x + phi_b()/360.))

    return p0, fitfunc, fitfunc_str


def fit_4gauss_fft(g_a, g_A, g_x0, g_sigma, g_B, g_x02, g_sigma2, g_C, g_x03, g_sigma3, g_D, g_x04, g_sigma4):
    fitfunc_str = 'a + A * exp(-(x-x0)**2/(2*sigma**2) + B * exp(-(x-x02)**2/(2*sigma2**2)) + C * exp(-(x-x03)**2/(2*sigma3**2)) + D * exp(-(x-x04)**2/(2*sigma4**2))'

    a = Parameter(g_a, 'a')
    x0 = Parameter(g_x0, 'x0')
    A = Parameter(g_A, 'A')
    sigma = Parameter(g_sigma, 'sigma')
    x02 = Parameter(g_x02, 'x02')
    B = Parameter(g_B, 'B')
    sigma2 = Parameter(g_sigma2, 'sigma2')
    x03 = Parameter(g_x03, 'x03')
    C = Parameter(g_C, 'C')
    sigma3 = Parameter(g_sigma3, 'sigma3')
    x04 = Parameter(g_x04, 'x04')
    D = Parameter(g_D, 'D')
    sigma4 = Parameter(g_sigma4, 'sigma4')

    p0 = [a, A, x0, sigma, B, x02, sigma2, C, x03, sigma3, D, x04, sigma4]

    def fitfunc(x):
        return a() + A() * np.exp(-(x-x0())**2/(2*sigma()**2)) + B() * np.exp(-(x-x02())**2/(2*sigma2()**2))  + C() * np.exp(-(x-x03())**2/(2*sigma3()**2)) + D() * np.exp(-(x-x04())**2/(2*sigma4()**2))
    return p0, fitfunc, fitfunc_str

def fit_2gauss_symmetric_fft(g_a, g_A, g_x0, g_sigma, g_B, g_x02, g_sigma2):
    fitfunc_str = 'a + A * exp(-(x-x0)**2/(2*sigma**2) + B * exp(-(x-x02)**2/(2*sigma2**2)) + A * exp(-(x+x0)**2/(2*sigma**2)) + B * exp(-(x+x02)**2/(2*sigma2**2))'

    a = Parameter(g_a, 'a')
    x0 = Parameter(g_x0, 'x0')
    A = Parameter(g_A, 'A')
    sigma = Parameter(g_sigma, 'sigma')
    x02 = Parameter(g_x02, 'x02')
    B = Parameter(g_B, 'B')
    sigma2 = Parameter(g_sigma2, 'sigma2')

    p0 = [a, A, x0, sigma, B, x02, sigma2]

    def fitfunc(x):
        return a() + A() * np.exp(-(x-x0())**2/(2*sigma()**2)) + B() * np.exp(-(x-x02())**2/(2*sigma2()**2))  + A() * np.exp(-(x+x0())**2/(2*sigma()**2)) + B() * np.exp(-(x+x02())**2/(2*sigma2()**2))
    return p0, fitfunc, fitfunc_str


def fit_2gauss_fft(g_a, g_A, g_x0, g_sigma, g_B, g_x02, g_sigma2):
    fitfunc_str = 'a + A * exp(-(x-x0)**2/(2*sigma**2) + B * exp(-(x-x02)**2/(2*sigma2**2))'

    a = Parameter(g_a, 'a')
    x0 = Parameter(g_x0, 'x0')
    A = Parameter(g_A, 'A')
    sigma = Parameter(g_sigma, 'sigma')
    x02 = Parameter(g_x02, 'x02')
    B = Parameter(g_B, 'B')
    sigma2 = Parameter(g_sigma2, 'sigma2')

    p0 = [a, A, x0, sigma, B, x02, sigma2]

    def fitfunc(x):
        return a() + A() * np.exp(-(x-x0())**2/(2*sigma()**2)) + B() * np.exp(-(x-x02())**2/(2*sigma2()**2))
    return p0, fitfunc, fitfunc_str

def fit_gauss_fft(g_a, g_A, g_x0, g_sigma):
    fitfunc_str = 'a + A * exp(-(x-x0)**2/(2*sigma**2)+ A * exp(-(x+x0)**2/(2*sigma**2))'

    a = Parameter(g_a, 'a')
    x0 = Parameter(g_x0, 'x0')
    A = Parameter(g_A, 'A')
    sigma = Parameter(g_sigma, 'sigma')

    p0 = [a, A, x0, sigma]

    def fitfunc(x):
        return a() + A() * np.exp(-(x-x0())**2/(2*sigma()**2))+ A() * np.exp(-(x+x0())**2/(2*sigma()**2))
    return p0, fitfunc, fitfunc_str


def fit_general_exponential_dec_cos(g_a, g_A, g_x0, g_T, g_n,g_f,g_phi):
    fitfunc_str = 'a + A * exp(-((x-x0)/T )**n*cos(2pi *(f*x+phi/360) )'

    a = Parameter(g_a, 'a')
    A = Parameter(g_A, 'A')
    x0 = Parameter(g_x0, 'x0')
    T = Parameter(g_T, 'T')
    n = Parameter(g_n, 'n')
    f = Parameter(g_f, 'f')
    phi = Parameter(g_phi, 'phi')


    p0 = [a, A, x0, T, n,f,phi]
    def fitfunc(x):
        return a() + A() * np.exp(-((x-x0())/T())**n())*np.cos(2*np.pi*( f()*x + phi()/360.))
    return p0, fitfunc, fitfunc_str


def fit_4lorentz_splitting(g_a1, g_A1, g_x01, g_gamma1, g_A2, g_dx2, g_gamma2, g_A3, g_x03, g_gamma3, g_A4, g_dx4, g_gamma4):
    fitfunc_str = 'a1 + 2*A1/np.pi*gamma1/(4*(x-x01)**2+gamma1**2) \
            + 2*A2/np.pi*gamma2/(4*(x-x01-dx2)**2+gamma2**2)  + 2*A3/np.pi*gamma3/(4*(x-x01-dx3)**2+gamma3**2) + 2*A4/np.pi*gamma4/(4*(x-x01-dx4)**2+gamma4**2)'

    a1 = Parameter(g_a1, 'a1')
    A1 = Parameter(g_A1, 'A1')
    x01 = Parameter(g_x01, 'x01')
    gamma1 = Parameter(g_gamma1, 'gamma1')

    dx2 = Parameter(g_dx2, 'dx2')
    A2 = Parameter(g_A2, 'A2')
    gamma2 = Parameter(g_gamma2, 'gamma2')

    x03 = Parameter(g_x03, 'x03')
    A3 = Parameter(g_A3, 'A3')
    gamma3 = Parameter(g_gamma3, 'gamma3')

    dx4 = Parameter(g_dx4, 'dx4')
    A4 = Parameter(g_A4, 'A4')
    gamma4 = Parameter(g_gamma4, 'gamma4')

    p0 = [a1, A1, x01, gamma1, A2, dx2, gamma2, A3, x03, gamma3, A4, dx4, gamma4]

    def fitfunc(x):
        return a1()+2*A1()/np.pi*gamma1()/(4*(x-x01())**2+gamma1()**2)+\
                2*A2()/np.pi*gamma2()/(4*(x-(x01()+dx2()))**2+gamma2()**2)+\
                2*A3()/np.pi*gamma3()/(4*(x-(x03()))**2+gamma3()**2)+\
                2*A4()/np.pi*gamma4()/(4*(x-(x03()+dx4()))**2+gamma4()**2)

    return p0, fitfunc, fitfunc_str



def fit_general_decaying_2cos_different_amplitudes(g_avg, g_t, g_A, g_f_a, g_phi_a, g_B, g_f_b, g_phi_b, g_n, *arg):
    fitfunc_str = 'A *exp(-x/t)**n( cos(2pi * (f1*x + phi1/360) )+ cos(2pi * (f2*x + phi2/360))) + a'

    f_a = Parameter(g_f_a, 'f_a')
    f_b = Parameter(g_f_b, 'f_b')
    a = Parameter(g_avg, 'a')
    A = Parameter(g_A, 'A')
    B = Parameter(g_B, 'B')
    phi_a = Parameter(g_phi_a, 'phi_a')
    phi_b = Parameter(g_phi_b, 'phi_b')
    t   = Parameter(g_t, 't')
    n = Parameter(g_n, 'n')
    p0 = [f_a, f_b, a, A, B, phi_a, phi_b, t, n]

    def fitfunc(x):
        return a() + np.exp(-(x/t())**n()) * (A()* np.cos(2*np.pi*( f_a()*x + phi_a()/360.))+B()*np.cos(2*np.pi*( f_b()*x + phi_b()/360.)))

    return p0, fitfunc, fitfunc_str